import asyncio
import time
import math
import pytest
from unittest import mock

from aioredis import ReplyError


@asyncio.coroutine
def add(redis, key, value):
    ok = yield from redis.connection.execute('set', key, value)
    assert ok == b'OK'


@pytest.mark.run_loop
def test_delete(redis):
    yield from add(redis, 'my-key', 123)
    yield from add(redis, 'other-key', 123)

    res = yield from redis.delete('my-key', 'non-existent-key')
    assert res == 1

    res = yield from redis.delete('other-key', 'other-key')
    assert res == 1

    with pytest.raises(TypeError):
        yield from redis.delete(None)

    with pytest.raises(TypeError):
        yield from redis.delete('my-key', 'my-key', None)


@pytest.mark.run_loop
def test_dump(redis):
    yield from add(redis, 'my-key', 123)

    data = yield from redis.dump('my-key')
    assert data == mock.ANY
    assert isinstance(data, (bytes, bytearray))
    assert len(data) > 0

    data = yield from redis.dump('non-existent-key')
    assert data is None

    with pytest.raises(TypeError):
        yield from redis.dump(None)


@pytest.mark.run_loop
def test_exists(redis, server):
    yield from add(redis, 'my-key', 123)

    res = yield from redis.exists('my-key')
    assert isinstance(res, int)
    assert res == 1

    res = yield from redis.exists('non-existent-key')
    assert isinstance(res, int)
    assert res == 0

    with pytest.raises(TypeError):
        yield from redis.exists(None)
    if server.version < (3, 0, 3):
        with pytest.raises(ReplyError):
            yield from redis.exists('key-1', 'key-2')


@pytest.redis_version(
    3, 0, 3, reason='Multi-key EXISTS available since redis>=2.8.0')
@pytest.mark.run_loop
def test_exists_multiple(redis):
    yield from add(redis, 'my-key', 123)

    res = yield from redis.exists('my-key', 'other-key')
    assert isinstance(res, int)
    assert res == 1

    res = yield from redis.exists('my-key', 'my-key')
    assert isinstance(res, int)
    assert res == 2

    res = yield from redis.exists('foo', 'bar')
    assert isinstance(res, int)
    assert res == 0


@pytest.mark.run_loop
def test_expire(redis):
    yield from add(redis, 'my-key', 132)

    res = yield from redis.expire('my-key', 10)
    assert res is True

    res = yield from redis.connection.execute('TTL', 'my-key')
    assert res >= 10

    yield from redis.expire('my-key', -1)
    res = yield from redis.exists('my-key')
    assert not res

    res = yield from redis.expire('other-key', 1000)
    assert res is False

    yield from add(redis, 'my-key', 1)
    res = yield from redis.expire('my-key', 10.0)
    assert res is True
    res = yield from redis.connection.execute('TTL', 'my-key')
    assert res >= 10

    with pytest.raises(TypeError):
        yield from redis.expire(None, 123)
    with pytest.raises(TypeError):
        yield from redis.expire('my-key', 'timeout')

# @pytest.mark.run_loop
# def test_wait_expire():
#     return
#     yield from .add('my-key', 123)
#     res = yield from .redis.expire('my-key', 1)
#     .assertIs(res, True)

#     yield from asyncio.sleep(1, loop=.loop)

#     res = yield from .redis.exists('my-key')
#     .assertIs(res, False)

# @pytest.mark.run_loop
# def test_wait_expireat():
#     return
#     yield from .add('my-key', 123)
#     ts = int(time.time() + 1)
#     res = yield from .redis.expireat('my-key', ts)

#     yield from asyncio.sleep(ts - time.time(), loop=.loop)
#     res = yield from .redis.exists('my-key')
#     .assertIs(res, False)


@pytest.mark.run_loop
def test_expireat(redis):
    yield from add(redis, 'my-key', 123)
    now = math.ceil(time.time())

    res = yield from redis.expireat('my-key', now + 10)
    assert res is True

    res = yield from redis.connection.execute('TTL', 'my-key')
    assert res >= 10

    res = yield from redis.expireat('my-key', -1)
    assert res is True

    res = yield from redis.exists('my-key')
    assert not res

    yield from add(redis, 'my-key', 123)

    res = yield from redis.expireat('my-key', 0)
    assert res is True

    res = yield from redis.exists('my-key')
    assert not res

    yield from add(redis, 'my-key', 123)
    res = yield from redis.expireat('my-key', time.time() + 10)
    assert res is True

    res = yield from redis.connection.execute('TTL', 'my-key')
    assert res >= 10

    yield from add(redis, 'my-key', 123)
    with pytest.raises(TypeError):
        yield from redis.expireat(None, 123)
    with pytest.raises(TypeError):
        yield from redis.expireat('my-key', 'timestamp')


@pytest.mark.run_loop
def test_keys(redis):
    res = yield from redis.keys('*pattern*')
    assert res == []

    yield from redis.connection.execute('FLUSHDB')
    res = yield from redis.keys('*')
    assert res == []

    yield from add(redis, 'my-key-1', 1)
    yield from add(redis, 'my-key-ab', 1)

    res = yield from redis.keys('my-key-?')
    assert res == [b'my-key-1']
    res = yield from redis.keys('my-key-*')
    assert sorted(res) == [b'my-key-1', b'my-key-ab']

    # test with encoding param
    res = yield from redis.keys('my-key-*', encoding='utf-8')
    assert sorted(res) == ['my-key-1', 'my-key-ab']

    with pytest.raises(TypeError):
        yield from redis.keys(None)


@pytest.mark.run_loop
def test_migrate(redis, create_redis, loop, serverB):
    yield from add(redis, 'my-key', 123)

    conn2 = yield from create_redis(serverB.tcp_address,
                                    db=2, loop=loop)
    yield from conn2.delete('my-key')
    assert (yield from redis.exists('my-key'))
    assert not (yield from conn2.exists('my-key'))

    ok = yield from redis.migrate(
        'localhost', serverB.tcp_address.port, 'my-key', 2, 1000)
    assert ok is True
    assert not (yield from redis.exists('my-key'))
    assert (yield from conn2.exists('my-key'))

    with pytest.raises_regex(TypeError, "host .* str"):
        yield from redis.migrate(None, 1234, 'key', 1, 23)
    with pytest.raises_regex(TypeError, "args .* None"):
        yield from redis.migrate('host', '1234',  None, 1, 123)
    with pytest.raises_regex(TypeError, "dest_db .* int"):
        yield from redis.migrate('host', 123, 'key', 1.0, 123)
    with pytest.raises_regex(TypeError, "timeout .* int"):
        yield from redis.migrate('host', '1234', 'key', 2, None)
    with pytest.raises_regex(ValueError, "Got empty host"):
        yield from redis.migrate('', '123', 'key', 1, 123)
    with pytest.raises_regex(ValueError, "dest_db .* greater equal 0"):
        yield from redis.migrate('host', 6379, 'key', -1, 1000)
    with pytest.raises_regex(ValueError, "timeout .* greater equal 0"):
        yield from redis.migrate('host', 6379, 'key', 1, -1000)


@pytest.mark.run_loop
def test_move(redis):
    yield from add(redis, 'my-key', 123)

    assert redis.db == 0
    res = yield from redis.move('my-key', 1)
    assert res is True

    with pytest.raises(TypeError):
        yield from redis.move(None, 1)
    with pytest.raises(TypeError):
        yield from redis.move('my-key', None)
    with pytest.raises(ValueError):
        yield from redis.move('my-key', -1)
    with pytest.raises(TypeError):
        yield from redis.move('my-key', 'not db')


@pytest.mark.run_loop
def test_object_refcount(redis):
    yield from add(redis, 'foo', 'bar')

    res = yield from redis.object_refcount('foo')
    assert res == 1
    res = yield from redis.object_refcount('non-existent-key')
    assert res is None

    with pytest.raises(TypeError):
        yield from redis.object_refcount(None)


@pytest.mark.run_loop
def test_object_encoding(redis, server):
    yield from add(redis, 'foo', 'bar')

    res = yield from redis.object_encoding('foo')

    if server.version < (3, 0, 0):
        assert res == b'raw'
    else:
        assert res == b'embstr'

    res = yield from redis.incr('key')
    assert res == 1
    res = yield from redis.object_encoding('key')
    assert res == b'int'
    res = yield from redis.object_encoding('non-existent-key')
    assert res is None

    with pytest.raises(TypeError):
        yield from redis.object_encoding(None)


@pytest.mark.run_loop
def test_object_idletime(redis, loop, server):
    yield from add(redis, 'foo', 'bar')

    res = yield from redis.object_idletime('foo')
    assert res == 0

    if server.version < (2, 8, 0):
        # Redis at least 2.6.x requires more time to sleep to incr idletime
        yield from asyncio.sleep(10, loop=loop)
    else:
        yield from asyncio.sleep(1, loop=loop)

    res = yield from redis.object_idletime('foo')
    assert res >= 1

    res = yield from redis.object_idletime('non-existent-key')
    assert res is None

    with pytest.raises(TypeError):
        yield from redis.object_idletime(None)


@pytest.mark.run_loop
def test_persist(redis):
    yield from add(redis, 'my-key', 123)
    res = yield from redis.expire('my-key', 10)
    assert res is True

    res = yield from redis.persist('my-key')
    assert res is True

    res = yield from redis.connection.execute('TTL', 'my-key')
    assert res == -1

    with pytest.raises(TypeError):
        yield from redis.persist(None)


@pytest.mark.run_loop
def test_pexpire(redis, loop):
    yield from add(redis, 'my-key', 123)
    res = yield from redis.pexpire('my-key', 100)
    assert res is True

    res = yield from redis.connection.execute('TTL', 'my-key')
    assert res == 0
    res = yield from redis.connection.execute('PTTL', 'my-key')
    assert res > 0

    yield from add(redis, 'my-key', 123)
    res = yield from redis.pexpire('my-key', 1)
    assert res is True

    yield from asyncio.sleep(.002, loop=loop)

    res = yield from redis.exists('my-key')
    assert not res

    with pytest.raises(TypeError):
        yield from redis.pexpire(None, 0)
    with pytest.raises(TypeError):
        yield from redis.pexpire('my-key', 1.0)


@pytest.mark.run_loop
def test_pexpireat(redis):
    yield from add(redis, 'my-key', 123)
    now = math.ceil((yield from redis.time()) * 1000)
    res = yield from redis.pexpireat('my-key', now + 800)
    assert res is True

    res = yield from redis.ttl('my-key')
    assert res == 1
    res = yield from redis.pttl('my-key')
    pytest.assert_almost_equal(res, 800, -2)

    with pytest.raises(TypeError):
        yield from redis.pexpireat(None, 1234)
    with pytest.raises(TypeError):
        yield from redis.pexpireat('key', 'timestamp')
    with pytest.raises(TypeError):
        yield from redis.pexpireat('key', 1000.0)


@pytest.mark.run_loop
def test_pttl(redis, server):
    yield from add(redis, 'key', 'val')
    res = yield from redis.pttl('key')
    assert res == -1
    res = yield from redis.pttl('non-existent-key')
    if server.version < (2, 8, 0):
        assert res == -1
    else:
        assert res == -2

    yield from redis.pexpire('key', 500)
    res = yield from redis.pttl('key')
    pytest.assert_almost_equal(res, 500, -2)

    with pytest.raises(TypeError):
        yield from redis.pttl(None)


@pytest.mark.run_loop
def test_randomkey(redis):
    yield from add(redis, 'key:1', 123)
    yield from add(redis, 'key:2', 123)
    yield from add(redis, 'key:3', 123)

    res = yield from redis.randomkey()
    assert res in [b'key:1', b'key:2', b'key:3']

    # test with encoding param
    res = yield from redis.randomkey(encoding='utf-8')
    assert res in ['key:1', 'key:2', 'key:3']

    yield from redis.connection.execute('flushdb')
    res = yield from redis.randomkey()
    assert res is None


@pytest.mark.run_loop
def test_rename(redis, server):
    yield from add(redis, 'foo', 'bar')
    yield from redis.delete('bar')

    res = yield from redis.rename('foo', 'bar')
    assert res is True

    with pytest.raises_regex(ReplyError, 'ERR no such key'):
        yield from redis.rename('foo', 'bar')
    with pytest.raises(TypeError):
        yield from redis.rename(None, 'bar')
    with pytest.raises(TypeError):
        yield from redis.rename('foo', None)
    with pytest.raises(ValueError):
        yield from redis.rename('foo', 'foo')

    if server.version < (3, 2):
        with pytest.raises_regex(ReplyError, '.* objects are the same'):
            yield from redis.rename('bar', b'bar')


@pytest.mark.run_loop
def test_renamenx(redis, server):
    yield from redis.delete('foo', 'bar')
    yield from add(redis, 'foo', 123)

    res = yield from redis.renamenx('foo', 'bar')
    assert res is True

    yield from add(redis, 'foo', 123)
    res = yield from redis.renamenx('foo', 'bar')
    assert res is False

    with pytest.raises_regex(ReplyError, 'ERR no such key'):
        yield from redis.renamenx('baz', 'foo')
    with pytest.raises(TypeError):
        yield from redis.renamenx(None, 'foo')
    with pytest.raises(TypeError):
        yield from redis.renamenx('foo', None)
    with pytest.raises(ValueError):
        yield from redis.renamenx('foo', 'foo')

    if server.version < (3, 2):
        with pytest.raises_regex(ReplyError, '.* objects are the same'):
            yield from redis.renamenx('foo', b'foo')


# @pytest.mark.run_loop
@pytest.mark.skip
def test_restore():
    pass


@pytest.redis_version(2, 8, 0, reason='SCAN is available since redis>=2.8.0')
@pytest.mark.run_loop
def test_scan(redis):
    for i in range(1, 11):
        foo_or_bar = 'bar' if i % 3 else 'foo'
        key = 'key:scan:{}:{}'.format(foo_or_bar, i).encode('utf-8')
        yield from add(redis, key, i)

    cursor, values = yield from redis.scan()
    # values should be *>=* just in case some other tests left
    # test keys
    assert len(values) >= 10

    cursor, test_values = b'0', []
    while cursor:
        cursor, values = yield from redis.scan(
            cursor=cursor, match=b'key:scan:foo*')
        test_values.extend(values)
    assert len(test_values) == 3

    cursor, test_values = b'0', []
    while cursor:
        cursor, values = yield from redis.scan(
            cursor=cursor, match=b'key:scan:bar:*')
        test_values.extend(values)
    assert len(test_values) == 7

    # SCAN family functions do not guarantee that the number of
    # elements returned per call are in a given range. So here
    # just dummy test, that *count* argument does not break something
    cursor = b'0'
    test_values = []
    while cursor:
        cursor, values = yield from redis.scan(cursor=cursor,
                                               match=b'key:scan:*',
                                               count=2)

        test_values.extend(values)
    assert len(test_values) == 10


@pytest.mark.run_loop
def test_sort(redis):
    def _make_list(key, items):
        yield from redis.delete(key)
        for i in items:
            yield from redis.rpush(key, i)

    yield from _make_list('a', '4231')
    res = yield from redis.sort('a')
    assert res == [b'1', b'2', b'3', b'4']

    res = yield from redis.sort('a', offset=2, count=2)
    assert res == [b'3', b'4']

    res = yield from redis.sort('a', asc=b'DESC')
    assert res == [b'4', b'3', b'2', b'1']

    yield from _make_list('a', 'dbca')
    res = yield from redis.sort(
        'a', asc=b'DESC', alpha=True, offset=2, count=2
    )
    assert res == [b'b', b'a']

    yield from redis.set('key:1', 10)
    yield from redis.set('key:2', 4)
    yield from redis.set('key:3', 7)
    yield from _make_list('a', '321')

    res = yield from redis.sort('a', by='key:*')
    assert res == [b'2', b'3', b'1']

    res = yield from redis.sort('a', by='nosort')
    assert res == [b'3', b'2', b'1']

    res = yield from redis.sort('a', by='key:*', store='sorted_a')
    assert res == 3
    res = yield from redis.lrange('sorted_a', 0, -1)
    assert res == [b'2', b'3', b'1']

    yield from redis.set('value:1', 20)
    yield from redis.set('value:2', 30)
    yield from redis.set('value:3', 40)
    res = yield from redis.sort('a', 'value:*', by='key:*')
    assert res == [b'30', b'40', b'20']

    yield from redis.hset('data_1', 'weight', 30)
    yield from redis.hset('data_2', 'weight', 20)
    yield from redis.hset('data_3', 'weight', 10)
    yield from redis.hset('hash_1', 'field', 20)
    yield from redis.hset('hash_2', 'field', 30)
    yield from redis.hset('hash_3', 'field', 10)
    res = yield from redis.sort(
        'a', 'hash_*->field', by='data_*->weight'
    )
    assert res == [b'10', b'30', b'20']


@pytest.mark.run_loop
def test_ttl(redis, server):
    yield from add(redis, 'key', 'val')
    res = yield from redis.ttl('key')
    assert res == -1
    res = yield from redis.ttl('non-existent-key')
    if server.version < (2, 8, 0):
        assert res == -1
    else:
        assert res == -2

    yield from redis.expire('key', 10)
    res = yield from redis.ttl('key')
    assert res >= 9

    with pytest.raises(TypeError):
        yield from redis.ttl(None)


@pytest.mark.run_loop
def test_type(redis):
    yield from add(redis, 'key', 'val')
    res = yield from redis.type('key')
    assert res == b'string'

    yield from redis.delete('key')
    yield from redis.incr('key')
    res = yield from redis.type('key')
    assert res == b'string'

    yield from redis.delete('key')
    yield from redis.sadd('key', 'val')
    res = yield from redis.type('key')
    assert res == b'set'

    res = yield from redis.type('non-existent-key')
    assert res == b'none'

    with pytest.raises(TypeError):
        yield from redis.type(None)



import time
import pytest
import sys

from unittest import mock

from aioredis import ReplyError


@pytest.mark.run_loop
def test_client_list(redis, server):
    res = yield from redis.client_list()
    assert isinstance(res, list)
    res = [dict(i._asdict()) for i in res]
    expected = {
        'addr': mock.ANY,
        'fd': mock.ANY,
        'age': '0',
        'idle': '0',
        'flags': 'N',
        'db': '0',
        'sub': '0',
        'psub': '0',
        'multi': '-1',
        'qbuf': '0',
        'qbuf_free': mock.ANY,
        'obl': '0',
        'oll': '0',
        'omem': '0',
        'events': 'r',
        'cmd': 'client',
        'name': '',
        }
    if server.version >= (2, 8, 12):
        expected['id'] = mock.ANY
    assert res == [expected]


@pytest.mark.run_loop
@pytest.mark.skipif(sys.platform == 'win32',
                    reason="No unixsocket on Windows")
def test_client_list__unixsocket(create_redis, loop, server):
    redis = yield from create_redis(server.unixsocket, loop=loop)
    res = yield from redis.client_list()
    assert len(res) == 1
    info = dict(res[0]._asdict())
    expected = {
        'addr': '{}:0'.format(server.unixsocket),
        'fd': mock.ANY,
        'age': '0',
        'idle': '0',
        'flags': 'U',   # Conneted via unix socket
        'db': '0',
        'sub': '0',
        'psub': '0',
        'multi': '-1',
        'qbuf': '0',
        'qbuf_free': mock.ANY,
        'obl': '0',
        'oll': '0',
        'omem': '0',
        'events': 'r',
        'cmd': 'client',
        'name': '',
        }
    if server.version >= (2, 8, 12):
        expected['id'] = mock.ANY
    assert info == expected


@pytest.mark.run_loop
@pytest.redis_version(
    2, 9, 50, reason='CLIENT PAUSE is available since redis >= 2.9.50')
def test_client_pause(redis):
    res = yield from redis.client_pause(2000)
    assert res is True
    ts = time.time()
    yield from redis.ping()
    dt = int(time.time() - ts)
    assert dt == 2

    with pytest.raises(TypeError):
        yield from redis.client_pause(2.0)
    with pytest.raises(ValueError):
        yield from redis.client_pause(-1)


@pytest.mark.run_loop
def test_client_getname(redis):
    res = yield from redis.client_getname()
    assert res is None
    ok = yield from redis.client_setname('TestClient')
    assert ok is True

    res = yield from redis.client_getname()
    assert res == b'TestClient'
    res = yield from redis.client_getname(encoding='utf-8')
    assert res == 'TestClient'


@pytest.mark.run_loop
def test_config_get(redis, server):
    res = yield from redis.config_get('port')
    assert res == {'port': str(server.tcp_address.port)}

    res = yield from redis.config_get()
    assert len(res) > 0

    res = yield from redis.config_get('unknown_parameter')
    assert res == {}

    with pytest.raises(TypeError):
        yield from redis.config_get(b'port')


@pytest.mark.run_loop
def test_config_rewrite(redis):
    with pytest.raises(ReplyError):
        yield from redis.config_rewrite()


@pytest.mark.run_loop
def test_config_set(redis):
    cur_value = yield from redis.config_get('slave-read-only')
    res = yield from redis.config_set('slave-read-only', 'no')
    assert res is True
    res = yield from redis.config_set(
        'slave-read-only', cur_value['slave-read-only'])
    assert res is True

    with pytest.raises_regex(ReplyError, "Unsupported CONFIG parameter"):
        yield from redis.config_set('databases', 100)
    with pytest.raises(TypeError):
        yield from redis.config_set(100, 'databases')


@pytest.mark.run_loop
@pytest.mark.skip("Not implemented")
def test_config_resetstat():
    pass


@pytest.mark.run_loop
def test_dbsize(redis):
    res = yield from redis.dbsize()
    assert res == 0

    yield from redis.set('key', 'value')

    res = yield from redis.dbsize()
    assert res > 0

    yield from redis.flushdb()
    res = yield from redis.dbsize()
    assert res == 0
    yield from redis.set('key', 'value')
    res = yield from redis.dbsize()
    assert res == 1


@pytest.mark.run_loop
def test_info(redis):
    res = yield from redis.info()
    assert isinstance(res, dict)

    res = yield from redis.info('all')
    assert isinstance(res, dict)

    with pytest.raises(ValueError):
        yield from redis.info('')


@pytest.mark.run_loop
@pytest.redis_version(2, 8, 12, reason='ROLE is available since redis>=2.8.12')
def test_role(redis):
    res = yield from redis.role()
    assert dict(res._asdict()) == {
        'role': 'master',
        'replication_offset': mock.ANY,
        'slaves': [],
        }


@pytest.mark.run_loop
def test_time(redis):
    res = yield from redis.time()
    assert isinstance(res, float)
    pytest.assert_almost_equal(int(res), int(time.time()), delta=10)


@pytest.mark.run_loop
def test_time_with_encoding(create_redis, server, loop):
    redis = yield from create_redis(server.tcp_address, loop=loop,
                                    encoding='utf-8')
    res = yield from redis.time()
    assert isinstance(res, float)
    pytest.assert_almost_equal(int(res), int(time.time()), delta=10)



import pytest

from aioredis import create_reconnecting_redis


@pytest.mark.run_loop
def test_recon(server, loop):
    redis = yield from create_reconnecting_redis(
        server.tcp_address, db=1, loop=loop)
    assert repr(redis) == '<Redis <AutoConnector None>>'
    resp = yield from redis.echo('ECHO')
    assert resp == b'ECHO'
    assert repr(redis) == '<Redis <AutoConnector <RedisConnection [db:1]>>>'
    conn_id = id(redis._conn._conn)

    redis._conn._conn._do_close(ValueError("Emulate connection close"))

    resp = yield from redis.echo('ECHO')
    assert resp == b'ECHO'
    assert conn_id != id(redis._conn._conn)
    # FIXME: bad interface
    conn = yield from redis.connection.get_atomic_connection()
    conn.close()
    yield from conn.wait_closed()


@pytest.mark.run_loop
def test_multi_exec(server, loop):
    redis = yield from create_reconnecting_redis(
        server.tcp_address, db=1, loop=loop)
    assert repr(redis) == '<Redis <AutoConnector None>>'

    m = redis.multi_exec()
    m.echo('ECHO')
    res = yield from m.execute()
    assert res == [b'ECHO']
    # FIXME: bad interface
    conn = yield from redis.connection.get_atomic_connection()
    conn.close()
    yield from conn.wait_closed()



import asyncio
import pytest
import socket
import subprocess
import sys
import re
import contextlib
import os
import ssl
import time
import logging

from collections import namedtuple

import aioredis


# Public fixtures


@pytest.yield_fixture
def loop():
    """Creates new event loop."""
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(None)

    yield loop

    if hasattr(loop, 'is_closed'):
        closed = loop.is_closed()
    else:
        closed = loop._closed   # XXX
    if not closed:
        loop.call_soon(loop.stop)
        loop.run_forever()
        loop.close()


@pytest.fixture(scope='session')
def unused_port():
    """Gets random free port."""
    def fun():
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.bind(('127.0.0.1', 0))
            return s.getsockname()[1]
    return fun


@pytest.fixture
def create_connection(_closable):
    """Wrapper around aioredis.create_connection."""

    @asyncio.coroutine
    def f(*args, **kw):
        conn = yield from aioredis.create_connection(*args, **kw)
        _closable(conn)
        return conn
    return f


@pytest.fixture
def create_redis(_closable, loop):
    """Wrapper around aioredis.create_redis."""

    @asyncio.coroutine
    def f(*args, **kw):
        redis = yield from aioredis.create_redis(*args, **kw)
        _closable(redis)
        return redis
    return f


@pytest.fixture
def create_pool(_closable):
    """Wrapper around aioredis.create_pool."""

    @asyncio.coroutine
    def f(*args, **kw):
        redis = yield from aioredis.create_pool(*args, **kw)
        _closable(redis)
        return redis
    return f


@pytest.fixture
def pool(create_pool, server, loop):
    """Returns RedisPool instance."""
    pool = loop.run_until_complete(
        create_pool(server.tcp_address, loop=loop))
    return pool


@pytest.fixture
def redis(create_redis, server, loop):
    """Returns Redis client instance."""
    redis = loop.run_until_complete(
        create_redis(server.tcp_address, loop=loop))
    loop.run_until_complete(redis.flushall())
    return redis


@pytest.yield_fixture
def _closable(loop):
    conns = []

    yield conns.append

    waiters = []
    while conns:
        conn = conns.pop(0)
        conn.close()
        waiters.append(conn.wait_closed())
    if waiters:
        loop.run_until_complete(asyncio.gather(*waiters, loop=loop))


@pytest.fixture(scope='session')
def server(start_server):
    """Starts redis-server instance."""
    return start_server('A')


@pytest.fixture(scope='session')
def serverB(start_server):
    """Starts redis-server instance."""
    return start_server('B')


# Internal stuff #


def pytest_addoption(parser):
    parser.addoption('--redis-server', default=[],
                     action="append",
                     help="Path to redis-server executable,"
                          " defaults to `%(default)s`")
    parser.addoption('--ssl-cafile', default='tests/ssl/test.crt',
                     help="Path to testing SSL certificate")
    parser.addoption('--uvloop', default=False,
                     action='store_true',
                     help="Run tests with uvloop")


def _read_server_version(redis_bin):
    args = [redis_bin, '--version']
    with subprocess.Popen(args, stdout=subprocess.PIPE) as proc:
        version = proc.stdout.readline().decode('utf-8')
    for part in version.split():
        if part.startswith('v='):
            break
    else:
        raise RuntimeError(
            "No version info can be found in {}".format(version))
    return tuple(map(int, part[2:].split('.')))


REDIS_SERVERS = []


@pytest.yield_fixture(scope='session', params=REDIS_SERVERS)
def start_server(request, unused_port):

    TCPAddress = namedtuple('TCPAddress', 'host port')
    RedisServer = namedtuple(
        'RedisServer', 'name tcp_address unixsocket version')
    processes = []
    servers = {}

    version = _read_server_version(request.param)
    verbose = request.config.getoption('-v') > 3

    def maker(name):
        if name in servers:
            return servers[name]

        port = unused_port()
        tcp_address = TCPAddress('localhost', port)
        cmd = [request.param,
               '--daemonize', 'no',
               '--save', '""',
               '--port', str(port),
               ]
        if sys.platform == 'win32':
            unixsocket = None
        else:
            unixsocket = '/tmp/redis.{}.sock'.format(port)
            cmd.extend(['--unixsocket', unixsocket])

        proc = subprocess.Popen(cmd,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT)
        processes.append(proc)
        log = b''
        while b'The server is now ready to accept connections ' not in log:
            assert proc.poll() is None, (
                "Process terminated", proc.returncode, proc.stdout.read())
            log = proc.stdout.readline()
            if log and verbose:
                print(log)

        info = RedisServer(name, tcp_address, unixsocket, version)
        servers.setdefault(name, info)
        return info

    yield maker

    while processes:
        proc = processes.pop(0)
        proc.terminate()
        proc.wait()


@pytest.yield_fixture(scope='session')
def ssl_proxy(request, unused_port):
    processes = []
    by_port = {}

    cafile = os.path.abspath(request.config.getoption('--ssl-cafile'))
    pemfile = os.path.splitext(cafile)[0] + '.pem'
    assert os.path.exists(cafile), \
        "No test ssl certificate, run `make certificate`"
    assert os.path.exists(pemfile), \
        "No test ssl certificate, run `make certificate`"

    if hasattr(ssl, 'create_default_context'):
        ssl_ctx = ssl.create_default_context(cafile=cafile)
    else:
        ssl_ctx = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
        ssl_ctx.load_verify_locations(cafile=cafile)
    if hasattr(ssl_ctx, 'check_hostname'):
        # available since python 3.4
        ssl_ctx.check_hostname = False
    ssl_ctx.verify_mode = ssl.CERT_NONE

    def sockat(unsecure_port):
        if unsecure_port in by_port:
            return by_port[unsecure_port]

        secure_port = unused_port()
        proc = subprocess.Popen(['/usr/bin/socat',
                                 'openssl-listen:{},'
                                 'cert={},'
                                 'verify=0,fork'
                                 .format(secure_port, pemfile),
                                 'tcp-connect:localhost:{}'
                                 .format(unsecure_port)
                                 ])
        processes.append(proc)
        time.sleep(1)   # XXX
        by_port[unsecure_port] = secure_port, ssl_ctx
        return secure_port, ssl_ctx

    yield sockat

    while processes:
        proc = processes.pop(0)
        proc.terminate()
        proc.wait()


@pytest.mark.tryfirst
def pytest_pycollect_makeitem(collector, name, obj):
    if collector.funcnamefilter(name):
        if not callable(obj):
            return
        item = pytest.Function(name, parent=collector)
        if 'run_loop' in item.keywords:
            # TODO: re-wrap with asyncio.coroutine if not native coroutine
            return list(collector._genfunctions(name, obj))


@pytest.mark.tryfirst
def pytest_pyfunc_call(pyfuncitem):
    """
    Run asyncio marked test functions in an event loop instead of a normal
    function call.
    """
    if 'run_loop' in pyfuncitem.keywords:
        funcargs = pyfuncitem.funcargs
        loop = funcargs['loop']
        testargs = {arg: funcargs[arg]
                    for arg in pyfuncitem._fixtureinfo.argnames}
        loop.run_until_complete(
            asyncio.wait_for(pyfuncitem.obj(**testargs),
                             15, loop=loop))
        return True


def pytest_runtest_setup(item):
    if 'run_loop' in item.keywords and 'loop' not in item.fixturenames:
        # inject an event loop fixture for all async tests
        item.fixturenames.append('loop')


def pytest_ignore_collect(path, config):
    if 'py35' in str(path):
        if sys.version_info < (3, 5, 0):
            return True


def pytest_collection_modifyitems(session, config, items):
    versions = {srv: _read_server_version(srv)
                for srv in REDIS_SERVERS}
    assert versions, ("Expected to detect redis versions", REDIS_SERVERS)
    for item in items:
        if 'redis_version' in item.keywords:
            marker = item.keywords['redis_version']
            kw = item.keywords
            version = [v for k, v in versions.items()
                       if k in kw or k.replace('\\', '\\\\') in kw]
            assert version, ("No version found", versions, item.keywords)
            version = version[0]
            if version < marker.kwargs['version']:
                item.add_marker(pytest.mark.skip(
                    reason=marker.kwargs['reason']))
        if 'ssl_proxy' in item.fixturenames:
            item.add_marker(pytest.mark.skipif(
                "not os.path.exists('/usr/bin/socat')",
                reason="socat package required (apt-get install socat)"))


def pytest_configure(config):
    bins = config.getoption('--redis-server')[:]
    REDIS_SERVERS[:] = bins or ['/usr/bin/redis-server']
    if config.getoption('--uvloop'):
        try:
            import uvloop
        except ImportError:
            raise RuntimeError(
                "Can not import uvloop, make sure it is installed")
        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())


@contextlib.contextmanager
def raises_regex(exc_type, message):
    with pytest.raises(exc_type) as exc_info:
        yield exc_info
    match = re.search(message, str(exc_info.value))
    assert match is not None, (
        "Pattern {!r} does not match {!r}"
        .format(message, str(exc_info.value)))


def logs(logger, level=None):
    """Catches logs for given logger and level.

    See unittest.TestCase.assertLogs for details.
    """
    return _AssertLogsContext(logger, level)


_LoggingWatcher = namedtuple("_LoggingWatcher", ["records", "output"])


class _CapturingHandler(logging.Handler):
    """
    A logging handler capturing all (raw and formatted) logging output.
    """

    def __init__(self):
        logging.Handler.__init__(self)
        self.watcher = _LoggingWatcher([], [])

    def flush(self):
        pass

    def emit(self, record):
        self.watcher.records.append(record)
        msg = self.format(record)
        self.watcher.output.append(msg)


class _AssertLogsContext:
    """Standard unittest's _AssertLogsContext context manager
    adopted to raise pytest failure.
    """
    LOGGING_FORMAT = "%(levelname)s:%(name)s:%(message)s"

    def __init__(self, logger_name, level):
        self.logger_name = logger_name
        if level:
            self.level = level
        else:
            self.level = logging.INFO
        self.msg = None

    def __enter__(self):
        if isinstance(self.logger_name, logging.Logger):
            logger = self.logger = self.logger_name
        else:
            logger = self.logger = logging.getLogger(self.logger_name)
        formatter = logging.Formatter(self.LOGGING_FORMAT)
        handler = _CapturingHandler()
        handler.setFormatter(formatter)
        self.watcher = handler.watcher
        self.old_handlers = logger.handlers[:]
        self.old_level = logger.level
        self.old_propagate = logger.propagate
        logger.handlers = [handler]
        logger.setLevel(self.level)
        logger.propagate = False
        return handler.watcher

    def __exit__(self, exc_type, exc_value, tb):
        self.logger.handlers = self.old_handlers
        self.logger.propagate = self.old_propagate
        self.logger.setLevel(self.old_level)
        if exc_type is not None:
            # let unexpected exceptions pass through
            return False
        if len(self.watcher.records) == 0:
            pytest.fail(
                "no logs of level {} or higher triggered on {}"
                .format(logging.getLevelName(self.level), self.logger.name))


def redis_version(*version, reason):
    assert 1 < len(version) <= 3, version
    assert all(isinstance(v, int) for v in version), version
    return pytest.mark.redis_version(version=version, reason=reason)


def assert_almost_equal(first, second, places=None, msg=None, delta=None):
    assert not (places is None and delta is None)
    if delta is not None:
        assert abs(first - second) <= delta
    else:
        assert round(abs(first - second), places) == 0


def pytest_namespace():
    return {
        'raises_regex': raises_regex,
        'assert_almost_equal': assert_almost_equal,
        'redis_version': redis_version,
        'logs': logs,
        }



import asyncio
import aioredis


def main():
    loop = asyncio.get_event_loop()

    async def go():
        pool = await aioredis.create_pool(
            ('localhost', 6379),
            minsize=5, maxsize=10)
        with await pool as redis:    # high-level redis API instance
            await redis.set('my-key', 'value')
            val = await redis.get('my-key')
        print('raw value:', val)
        pool.close()
        await pool.wait_closed()    # closing all open connections

    loop.run_until_complete(go())


if __name__ == '__main__':
    main()



import asyncio
import pytest

from unittest.mock import patch

from aioredis import (
    RedisPool,
    ReplyError,
    PoolClosedError,
    ConnectionClosedError,
    )


def _assert_defaults(pool):
    assert isinstance(pool, RedisPool)
    assert pool.minsize == 1
    assert pool.maxsize == 10
    assert pool.size == 1
    assert pool.freesize == 1
    assert pool._close_waiter is None


def test_connect(pool):
    _assert_defaults(pool)


def test_global_loop(create_pool, loop, server):
    asyncio.set_event_loop(loop)

    pool = loop.run_until_complete(create_pool(
        server.tcp_address))
    _assert_defaults(pool)


@pytest.mark.run_loop
def test_clear(pool):
    _assert_defaults(pool)

    yield from pool.clear()
    assert pool.freesize == 0


@pytest.mark.run_loop
@pytest.mark.parametrize('minsize', [None, -100, 0.0, 100])
def test_minsize(minsize, create_pool, loop, server):

    with pytest.raises(AssertionError):
        yield from create_pool(
            server.tcp_address,
            minsize=minsize, maxsize=10, loop=loop)


@pytest.mark.run_loop
@pytest.mark.parametrize('maxsize', [None, -100, 0.0, 1])
def test_maxsize(maxsize, create_pool, loop, server):

    with pytest.raises(AssertionError):
        yield from create_pool(
            server.tcp_address,
            minsize=2, maxsize=maxsize, loop=loop)


@pytest.mark.run_loop
def test_create_connection_timeout(create_pool, loop, server):
    with patch('asyncio.open_connection') as\
            open_conn_mock:
        open_conn_mock.side_effect = lambda *a, **kw: asyncio.sleep(0.2,
                                                                    loop=loop)
        with pytest.raises(asyncio.TimeoutError):
            yield from create_pool(
                server.tcp_address, loop=loop,
                create_connection_timeout=0.1)


def test_no_yield_from(pool):
    with pytest.raises(RuntimeError):
        with pool:
            pass    # pragma: no cover


@pytest.mark.run_loop
def test_simple_command(create_pool, loop, server):
    pool = yield from create_pool(
        server.tcp_address,
        minsize=10, loop=loop)

    with (yield from pool) as conn:
        msg = yield from conn.echo('hello')
        assert msg == b'hello'
        assert pool.size == 10
        assert pool.freesize == 9
    assert pool.size == 10
    assert pool.freesize == 10


@pytest.mark.run_loop
def test_create_new(create_pool, loop, server):
    pool = yield from create_pool(
        server.tcp_address,
        minsize=1, loop=loop)
    assert pool.size == 1
    assert pool.freesize == 1

    with (yield from pool):
        assert pool.size == 1
        assert pool.freesize == 0

        with (yield from pool):
            assert pool.size == 2
            assert pool.freesize == 0

    assert pool.size == 2
    assert pool.freesize == 2


@pytest.mark.run_loop
def test_create_constraints(create_pool, loop, server):
    pool = yield from create_pool(
        server.tcp_address,
        minsize=1, maxsize=1, loop=loop)
    assert pool.size == 1
    assert pool.freesize == 1

    with (yield from pool):
        assert pool.size == 1
        assert pool.freesize == 0

        with pytest.raises(asyncio.TimeoutError):
            yield from asyncio.wait_for(pool.acquire(),
                                        timeout=0.2,
                                        loop=loop)


@pytest.mark.run_loop
def test_create_no_minsize(create_pool, loop, server):
    pool = yield from create_pool(
        server.tcp_address,
        minsize=0, maxsize=1, loop=loop)
    assert pool.size == 0
    assert pool.freesize == 0

    with (yield from pool):
        assert pool.size == 1
        assert pool.freesize == 0

        with pytest.raises(asyncio.TimeoutError):
            yield from asyncio.wait_for(pool.acquire(),
                                        timeout=0.2,
                                        loop=loop)
    assert pool.size == 1
    assert pool.freesize == 1


@pytest.mark.run_loop
def test_release_closed(create_pool, loop, server):
    pool = yield from create_pool(
        server.tcp_address,
        minsize=1, loop=loop)
    assert pool.size == 1
    assert pool.freesize == 1

    with (yield from pool) as redis:
        redis.close()
        yield from redis.wait_closed()
    assert pool.size == 0
    assert pool.freesize == 0


@pytest.mark.run_loop
def test_release_bad_connection(create_pool, create_redis, loop, server):
    pool = yield from create_pool(
        server.tcp_address,
        loop=loop)
    conn = yield from pool.acquire()
    other_conn = yield from create_redis(
        server.tcp_address,
        loop=loop)
    with pytest.raises(AssertionError):
        pool.release(other_conn)

    pool.release(conn)
    other_conn.close()
    yield from other_conn.wait_closed()


@pytest.mark.run_loop
def test_select_db(create_pool, loop, server):
    pool = yield from create_pool(
        server.tcp_address,
        loop=loop)

    yield from pool.select(1)
    with (yield from pool) as redis:
        assert redis.db == 1


@pytest.mark.run_loop
def test_change_db(create_pool, loop, server):
    pool = yield from create_pool(
        server.tcp_address,
        minsize=1, db=0,
        loop=loop)
    assert pool.size == 1
    assert pool.freesize == 1

    with (yield from pool) as redis:
        yield from redis.select(1)
    assert pool.size == 0
    assert pool.freesize == 0

    with (yield from pool) as redis:
        assert pool.size == 1
        assert pool.freesize == 0

        yield from pool.select(1)
        assert pool.db == 1
        assert pool.size == 1
        assert pool.freesize == 0
    assert pool.size == 0
    assert pool.freesize == 0
    assert pool.db == 1


@pytest.mark.run_loop
def test_change_db_errors(create_pool, loop, server):
    pool = yield from create_pool(
        server.tcp_address,
        minsize=1, db=0,
        loop=loop)

    with pytest.raises(TypeError):
        yield from pool.select(None)
    assert pool.db == 0

    with (yield from pool):
        pass
    assert pool.size == 1
    assert pool.freesize == 1

    with pytest.raises(TypeError):
        yield from pool.select(None)
    assert pool.db == 0
    with pytest.raises(ValueError):
        yield from pool.select(-1)
    assert pool.db == 0
    with pytest.raises(ReplyError):
        yield from pool.select(100000)
    assert pool.db == 0


@pytest.mark.run_loop
def test_select_and_create(create_pool, loop, server):
    # trying to model situation when select and acquire
    # called simultaneously
    # but acquire freezes on _wait_select and
    # then continues with propper db
    @asyncio.coroutine
    def test():
        pool = yield from create_pool(
            server.tcp_address,
            minsize=1, db=0,
            loop=loop)
        db = 0
        while True:
            db = (db + 1) & 1
            _, conn = yield from asyncio.gather(pool.select(db),
                                                pool.acquire(),
                                                loop=loop)
            assert pool.db == db
            pool.release(conn)
            if conn.db == db:
                break
    yield from asyncio.wait_for(test(), 3, loop=loop)


@pytest.mark.run_loop
def test_response_decoding(create_pool, loop, server):
    pool = yield from create_pool(
        server.tcp_address,
        encoding='utf-8', loop=loop)

    assert pool.encoding == 'utf-8'
    with (yield from pool) as redis:
        yield from redis.set('key', 'value')
    with (yield from pool) as redis:
        res = yield from redis.get('key')
        assert res == 'value'


@pytest.mark.run_loop
def test_hgetall_response_decoding(create_pool, loop, server):
    pool = yield from create_pool(
        server.tcp_address,
        encoding='utf-8', loop=loop)

    assert pool.encoding == 'utf-8'
    with (yield from pool) as redis:
        yield from redis.delete('key1')
        yield from redis.hmset('key1', 'foo', 'bar')
        yield from redis.hmset('key1', 'baz', 'zap')
    with (yield from pool) as redis:
        res = yield from redis.hgetall('key1')
        assert res == {'foo': 'bar', 'baz': 'zap'}


@pytest.mark.run_loop
def test_crappy_multiexec(create_pool, loop, server):
    pool = yield from create_pool(
        server.tcp_address,
        encoding='utf-8', loop=loop,
        minsize=1, maxsize=1)

    with (yield from pool) as redis:
        yield from redis.set('abc', 'def')
        yield from redis.connection.execute('multi')
        yield from redis.set('abc', 'fgh')
    assert redis.closed is True
    with (yield from pool) as redis:
        value = yield from redis.get('abc')
    assert value == 'def'


@pytest.mark.run_loop
def test_pool_size_growth(create_pool, server, loop):
    pool = yield from create_pool(
        server.tcp_address,
        loop=loop,
        minsize=1, maxsize=1)

    done = set()
    tasks = []

    @asyncio.coroutine
    def task1(i):
        with (yield from pool):
            assert pool.size <= pool.maxsize
            assert pool.freesize == 0
            yield from asyncio.sleep(0.2, loop=loop)
            done.add(i)

    @asyncio.coroutine
    def task2():
        with (yield from pool):
            assert pool.size <= pool.maxsize
            assert pool.freesize >= 0
            assert done == {0, 1}

    for _ in range(2):
        tasks.append(asyncio.async(task1(_), loop=loop))
    tasks.append(asyncio.async(task2(), loop=loop))
    yield from asyncio.gather(*tasks, loop=loop)


@pytest.mark.run_loop
def test_pool_with_closed_connections(create_pool, server, loop):
    pool = yield from create_pool(
        server.tcp_address,
        loop=loop,
        minsize=1, maxsize=2)
    assert 1 == pool.freesize
    conn1 = pool._pool[0]
    conn1.close()
    assert conn1.closed is True
    assert 1 == pool.freesize
    with (yield from pool) as conn2:
        assert conn2.closed is False
        assert conn1 is not conn2


@pytest.mark.run_loop
def test_pool_close(create_pool, server, loop):
    pool = yield from create_pool(
        server.tcp_address, loop=loop)

    assert pool.closed is False

    with (yield from pool) as redis:
        assert (yield from redis.ping()) == b'PONG'

    pool.close()
    yield from pool.wait_closed()
    assert pool.closed is True

    with pytest.raises(PoolClosedError):
        with (yield from pool) as redis:
            assert (yield from redis.ping()) == b'PONG'


@pytest.mark.run_loop
def test_pool_close__used(create_pool, server, loop):
    pool = yield from create_pool(
        server.tcp_address, loop=loop)

    assert pool.closed is False

    with (yield from pool) as redis:
        pool.close()
        yield from pool.wait_closed()
        assert pool.closed is True

        with pytest.raises(ConnectionClosedError):
            yield from redis.ping()


@pytest.mark.run_loop
@pytest.redis_version(2, 8, 0, reason="maxclients config setting")
def test_pool_check_closed_when_exception(create_pool, create_redis,
                                          start_server, loop):
    server = start_server('server-small')
    redis = yield from create_redis(server.tcp_address, loop=loop)
    yield from redis.config_set('maxclients', 2)

    with pytest.logs('aioredis', 'DEBUG') as cm:
        with pytest.raises(Exception):
            yield from create_pool(address=tuple(server.tcp_address),
                                   minsize=2, loop=loop)

    assert len(cm.output) >= 3
    connect_msg = (
        "DEBUG:aioredis:Creating tcp connection"
        " to ('localhost', {})".format(server.tcp_address.port))
    assert cm.output[:2] == [connect_msg, connect_msg]
    assert cm.output[-1] == "DEBUG:aioredis:Closed 1 connections"



import asyncio
import aioredis


def main():
    loop = asyncio.get_event_loop()

    async def reader(ch):
        while (await ch.wait_message()):
            msg = await ch.get_json()
            print("Got Message:", msg)

    async def go():
        pub = await aioredis.create_redis(
            ('localhost', 6379))
        sub = await aioredis.create_redis(
            ('localhost', 6379))
        res = await sub.subscribe('chan:1')
        ch1 = res[0]

        tsk = asyncio.ensure_future(reader(ch1))

        res = await pub.publish_json('chan:1', ["Hello", "world"])
        assert res == 1

        await sub.unsubscribe('chan:1')
        await tsk
        sub.close()
        pub.close()

    loop.run_until_complete(go())


if __name__ == '__main__':
    main()



import asyncio
import aioredis


def main():
    loop = asyncio.get_event_loop()

    async def go():
        redis = await aioredis.create_redis(
            ('localhost', 6379))
        await redis.set('my-key', 'value')
        val = await redis.get('my-key')
        print(val)

        # gracefully closing underlying connection
        redis.close()
        await redis.wait_closed()
    loop.run_until_complete(go())


if __name__ == '__main__':
    main()



import asyncio

from aioredis.connection import create_connection
from aioredis.util import _NOTSET
from .generic import GenericCommandsMixin
from .string import StringCommandsMixin
from .hash import HashCommandsMixin
from .hyperloglog import HyperLogLogCommandsMixin
from .set import SetCommandsMixin
from .sorted_set import SortedSetCommandsMixin
from .transaction import TransactionsCommandsMixin, Pipeline, MultiExec
from .list import ListCommandsMixin
from .scripting import ScriptingCommandsMixin
from .server import ServerCommandsMixin
from .pubsub import PubSubCommandsMixin
from .cluster import ClusterCommandsMixin
from .geo import GeoCommandsMixin, GeoPoint, GeoMember

__all__ = [
    'create_redis', 'Redis',
    'Pipeline', 'MultiExec',
    'GeoPoint', 'GeoMember',
]


class AutoConnector(object):
    closed = False

    def __init__(self, *conn_args, **conn_kwargs):
        self._conn_args = conn_args
        self._conn_kwargs = conn_kwargs
        self._conn = None
        self._loop = conn_kwargs.get('loop')
        self._lock = asyncio.Lock(loop=self._loop)

    def __repr__(self):
        return '<AutoConnector {!r}>'.format(self._conn)

    @asyncio.coroutine
    def execute(self, *args, **kwargs):
        conn = yield from self.get_atomic_connection()
        return (yield from conn.execute(*args, **kwargs))

    @asyncio.coroutine
    def get_atomic_connection(self):
        if self._conn is None or self._conn.closed:
            with (yield from self._lock):
                if self._conn is None or self._conn.closed:
                    conn = yield from create_connection(
                        *self._conn_args, **self._conn_kwargs)
                    self._conn = conn
        return self._conn


class Redis(GenericCommandsMixin, StringCommandsMixin,
            HyperLogLogCommandsMixin, SetCommandsMixin,
            HashCommandsMixin, TransactionsCommandsMixin,
            SortedSetCommandsMixin, ListCommandsMixin,
            ScriptingCommandsMixin, ServerCommandsMixin,
            PubSubCommandsMixin, ClusterCommandsMixin,
            GeoCommandsMixin):
    """High-level Redis interface.

    Gathers in one place Redis commands implemented in mixins.

    For commands details see: http://redis.io/commands/#connection
    """

    def __init__(self, connection):
        self._conn = connection

    def __repr__(self):
        return '<Redis {!r}>'.format(self._conn)

    def close(self):
        self._conn.close()

    @asyncio.coroutine
    def wait_closed(self):
        yield from self._conn.wait_closed()

    @property
    def db(self):
        """Currently selected db index."""
        return self._conn.db

    @property
    def encoding(self):
        """Current set codec or None."""
        return self._conn.encoding

    @property
    def connection(self):
        """:class:`aioredis.RedisConnection` instance."""
        return self._conn

    @property
    def in_transaction(self):
        """Set to True when MULTI command was issued."""
        return self._conn.in_transaction

    @property
    def closed(self):
        """True if connection is closed."""
        return self._conn.closed

    def auth(self, password):
        """Authenticate to server.

        This method wraps call to :meth:`aioredis.RedisConnection.auth()`
        """
        return self._conn.auth(password)

    def echo(self, message, *, encoding=_NOTSET):
        """Echo the given string."""
        return self._conn.execute('ECHO', message, encoding=encoding)

    def ping(self, *, encoding=_NOTSET):
        """Ping the server."""
        return self._conn.execute('PING', encoding=encoding)

    def quit(self):
        """Close the connection."""
        return self._conn.execute('QUIT')

    def select(self, db):
        """Change the selected database for the current connection.

        This method wraps call to :meth:`aioredis.RedisConnection.select()`
        """
        return self._conn.select(db)


@asyncio.coroutine
def create_redis(address, *, db=None, password=None, ssl=None,
                 encoding=None, commands_factory=Redis,
                 timeout=None,
                 loop=None):
    """Creates high-level Redis interface.

    This function is a coroutine.
    """
    conn = yield from create_connection(address, db=db,
                                        password=password,
                                        ssl=ssl,
                                        encoding=encoding,
                                        timeout=timeout,
                                        loop=loop)
    return commands_factory(conn)


@asyncio.coroutine
def create_reconnecting_redis(address, *, db=None, password=None, ssl=None,
                              encoding=None, commands_factory=Redis,
                              loop=None):
    """Creates high-level Redis interface.

    This function is a coroutine.
    """
    # Note: this is not coroutine, but we may make it such. We may start
    # a first connection in it, or just resolve DNS. So let's keep it
    # coroutine for forward compatibility
    conn = AutoConnector(address,
                         db=db, password=password, ssl=ssl,
                         encoding=encoding, loop=loop)
    return commands_factory(conn)


# make pyflakes happy
(Pipeline, MultiExec)



import asyncio
import collections
import sys
import warnings

from .commands import create_redis, Redis
from .log import logger
from .util import async_task, _NOTSET
from .errors import PoolClosedError
from .locks import Lock


PY_35 = sys.version_info >= (3, 5)


@asyncio.coroutine
def create_pool(address, *, db=0, password=None, ssl=None, encoding=None,
                minsize=1, maxsize=10, commands_factory=_NOTSET,
                loop=None, create_connection_timeout=None):
    # FIXME: rewrite docstring
    """Creates Redis Pool.

    By default it creates pool of Redis instances, but it is
    also possible to create pool of plain connections by passing
    ``lambda conn: conn`` as commands_factory.

    *commands_factory* parameter is deprecated since v0.2.9

    All arguments are the same as for create_connection.

    Returns RedisPool instance.
    """
    if commands_factory is not _NOTSET:
        warnings.warn(
            "commands_factory argument is deprecated and will be removed!",
            warnings.DeprecationWarning)
    else:
        commands_factory = Redis

    pool = RedisPool(address, db, password, encoding,
                     minsize=minsize, maxsize=maxsize,
                     commands_factory=commands_factory,
                     create_connection_timeout=create_connection_timeout,
                     ssl=ssl, loop=loop)
    try:
        yield from pool._fill_free(override_min=False)
    except Exception as ex:
        pool.close()
        yield from pool.wait_closed()
        raise

    return pool


class RedisPool:
    """Redis connections pool.
    """

    def __init__(self, address, db=0, password=None, encoding=None,
                 *, minsize, maxsize, commands_factory, ssl=None,
                 create_connection_timeout=None, loop=None):
        assert isinstance(minsize, int) and minsize >= 0, (
            "minsize must be int >= 0", minsize, type(minsize))
        assert maxsize is not None, "Arbitrary pool size is disallowed."
        assert isinstance(maxsize, int) and maxsize > 0, (
            "maxsize must be int > 0", maxsize, type(maxsize))
        assert minsize <= maxsize, (
            "Invalid pool min/max sizes", minsize, maxsize)
        if loop is None:
            loop = asyncio.get_event_loop()
        self._address = address
        self._db = db
        self._password = password
        self._ssl = ssl
        self._encoding = encoding
        self._minsize = minsize
        self._factory = commands_factory
        self._create_connection_timeout = create_connection_timeout
        self._loop = loop
        self._pool = collections.deque(maxlen=maxsize)
        self._used = set()
        self._acquiring = 0
        self._cond = asyncio.Condition(lock=Lock(loop=loop), loop=loop)
        self._close_state = asyncio.Event(loop=loop)
        self._close_waiter = None

    @property
    def minsize(self):
        """Minimum pool size."""
        return self._minsize

    @property
    def maxsize(self):
        """Maximum pool size."""
        return self._pool.maxlen

    @property
    def size(self):
        """Current pool size."""
        return self.freesize + len(self._used) + self._acquiring

    @property
    def freesize(self):
        """Current number of free connections."""
        return len(self._pool)

    @asyncio.coroutine
    def clear(self):
        """Clear pool connections.

        Close and remove all free connections.
        """
        with (yield from self._cond):
            waiters = []
            while self._pool:
                conn = self._pool.popleft()
                conn.close()
                waiters.append(conn.wait_closed())
            yield from asyncio.gather(*waiters, loop=self._loop)

    @asyncio.coroutine
    def _do_close(self):
        yield from self._close_state.wait()
        with (yield from self._cond):
            assert not self._acquiring, self._acquiring
            waiters = []
            while self._pool:
                conn = self._pool.popleft()
                conn.close()
                waiters.append(conn.wait_closed())
            for conn in self._used:
                conn.close()
                waiters.append(conn.wait_closed())
            yield from asyncio.gather(*waiters, loop=self._loop)
            logger.debug("Closed %d connections", len(waiters))

    def close(self):
        """Close all free and in-progress connections and mark pool as closed.
        """
        if not self._close_state.is_set():
            self._close_waiter = async_task(self._do_close(), loop=self._loop)
            self._close_state.set()

    @property
    def closed(self):
        """True if pool is closed."""
        return self._close_state.is_set()

    @asyncio.coroutine
    def wait_closed(self):
        """Wait until pool gets closed."""
        yield from self._close_state.wait()
        assert self._close_waiter is not None
        yield from asyncio.shield(self._close_waiter, loop=self._loop)

    @property
    def db(self):
        """Currently selected db index."""
        return self._db

    @property
    def encoding(self):
        """Current set codec or None."""
        return self._encoding

    @asyncio.coroutine
    def select(self, db):
        """Changes db index for all free connections.

        All previously acquired connections will be closed when released.
        """
        with (yield from self._cond):
            for i in range(self.freesize):
                yield from self._pool[i].select(db)
            else:
                self._db = db

    @asyncio.coroutine
    def acquire(self):
        """Acquires a connection from free pool.

        Creates new connection if needed.
        """
        if self.closed:
            raise PoolClosedError("Pool is closed")
        with (yield from self._cond):
            if self.closed:
                raise PoolClosedError("Pool is closed")
            while True:
                yield from self._fill_free(override_min=True)
                if self.freesize:
                    conn = self._pool.popleft()
                    assert not conn.closed, conn
                    assert conn not in self._used, (conn, self._used)
                    self._used.add(conn)
                    return conn
                else:
                    yield from self._cond.wait()

    def release(self, conn):
        """Returns used connection back into pool.

        When returned connection has db index that differs from one in pool
        the connection will be closed and dropped.
        When queue of free connections is full the connection will be dropped.
        """
        assert conn in self._used, "Invalid connection, maybe from other pool"
        self._used.remove(conn)
        if not conn.closed:
            if conn.in_transaction:
                logger.warning(
                    "Connection %r is in transaction, closing it.", conn)
                conn.close()
            elif conn.in_pubsub:
                logger.warning(
                    "Connection %r is in subscribe mode, closing it.", conn)
                conn.close()
            elif conn.db == self.db:
                if self.maxsize and self.freesize < self.maxsize:
                    self._pool.append(conn)
                else:
                    # consider this connection as old and close it.
                    conn.close()
            else:
                conn.close()
        # FIXME: check event loop is not closed
        async_task(self._wakeup(), loop=self._loop)

    def _drop_closed(self):
        for i in range(self.freesize):
            conn = self._pool[0]
            if conn.closed:
                self._pool.popleft()
            else:
                self._pool.rotate(1)

    @asyncio.coroutine
    def _fill_free(self, *, override_min):
        # drop closed connections first
        self._drop_closed()
        while self.size < self.minsize:
            self._acquiring += 1
            try:
                conn = yield from self._create_new_connection()
                self._pool.append(conn)
            finally:
                self._acquiring -= 1
                # connection may be closed at yield point
                self._drop_closed()
        if self.freesize:
            return
        if override_min:
            while not self._pool and self.size < self.maxsize:
                self._acquiring += 1
                try:
                    conn = yield from self._create_new_connection()
                    self._pool.append(conn)
                finally:
                    self._acquiring -= 1
                    # connection may be closed at yield point
                    self._drop_closed()

    def _create_new_connection(self):
        return create_redis(self._address,
                            db=self._db,
                            password=self._password,
                            ssl=self._ssl,
                            encoding=self._encoding,
                            commands_factory=self._factory,
                            timeout=self._create_connection_timeout,
                            loop=self._loop)

    @asyncio.coroutine
    def _wakeup(self, closing_conn=None):
        with (yield from self._cond):
            self._cond.notify()
        if closing_conn is not None:
            yield from closing_conn.wait_closed()

    def __enter__(self):
        raise RuntimeError(
            "'yield from' should be used as a context manager expression")

    def __exit__(self, *args):
        pass    # pragma: nocover

    def __iter__(self):
        # this method is needed to allow `yield`ing from pool
        conn = yield from self.acquire()
        return _ConnectionContextManager(self, conn)

    if PY_35:
        def __await__(self):
            # To make `with await pool` work
            conn = yield from self.acquire()
            return _ConnectionContextManager(self, conn)

        def get(self):
            '''Return async context manager for working with connection.

            async with pool.get() as conn:
                await conn.get(key)
            '''
            return _AsyncConnectionContextManager(self)


class _ConnectionContextManager:

    __slots__ = ('_pool', '_conn')

    def __init__(self, pool, conn):
        self._pool = pool
        self._conn = conn

    def __enter__(self):
        return self._conn

    def __exit__(self, exc_type, exc_value, tb):
        try:
            self._pool.release(self._conn)
        finally:
            self._pool = None
            self._conn = None


if PY_35:
    class _AsyncConnectionContextManager:

        __slots__ = ('_pool', '_conn')

        def __init__(self, pool):
            self._pool = pool
            self._conn = None

        @asyncio.coroutine
        def __aenter__(self):
            self._conn = yield from self._pool.acquire()
            return self._conn

        @asyncio.coroutine
        def __aexit__(self, exc_type, exc_value, tb):
            try:
                self._pool.release(self._conn)
            finally:
                self._pool = None
                self._conn = None



import asyncio
import pytest

from aioredis.pubsub import Receiver


@pytest.mark.run_loop
async def test_pubsub_receiver_iter(create_redis, server, loop):
    sub = await create_redis(server.tcp_address, loop=loop)
    pub = await create_redis(server.tcp_address, loop=loop)

    mpsc = Receiver(loop=loop)

    async def coro(mpsc):
        lst = []
        async for msg in mpsc.iter():
            lst.append(msg)
        return lst

    tsk = asyncio.ensure_future(coro(mpsc), loop=loop)
    snd1, = await sub.subscribe(mpsc.channel('chan:1'))
    snd2, = await sub.subscribe(mpsc.channel('chan:2'))
    snd3, = await sub.psubscribe(mpsc.pattern('chan:*'))

    await pub.publish_json('chan:1', {'Hello': 'World'})
    await pub.publish_json('chan:2', ['message'])
    mpsc.stop()
    await asyncio.sleep(0, loop=loop)
    assert await tsk == [
        (snd1, b'{"Hello": "World"}'),
        (snd3, (b'chan:1', b'{"Hello": "World"}')),
        (snd2, b'["message"]'),
        (snd3, (b'chan:2', b'["message"]')),
        ]
    assert not mpsc.is_active



import pytest
import sys

from aioredis import ConnectionClosedError, ReplyError


@pytest.mark.run_loop
def test_repr(create_redis, loop, server):
    redis = yield from create_redis(
        server.tcp_address, db=1, loop=loop)
    assert repr(redis) == '<Redis <RedisConnection [db:1]>>'

    redis = yield from create_redis(
        server.tcp_address, db=0, loop=loop)
    assert repr(redis) == '<Redis <RedisConnection [db:0]>>'


@pytest.mark.run_loop
def test_auth(redis):
    expected_message = "ERR Client sent AUTH, but no password is set"
    with pytest.raises_regex(ReplyError, expected_message):
        yield from redis.auth('')


@pytest.mark.run_loop
def test_echo(redis):
    resp = yield from redis.echo('ECHO')
    assert resp == b'ECHO'

    with pytest.raises(TypeError):
        yield from redis.echo(None)


@pytest.mark.run_loop
def test_ping(redis):
    resp = yield from redis.ping()
    assert resp == b'PONG'


@pytest.mark.xfail(sys.platform == 'win32',
                   reason="Probably race conditions...")
@pytest.mark.run_loop
def test_quit(redis):
    resp = yield from redis.quit()
    assert resp == b'OK'

    with pytest.raises(ConnectionClosedError):
        yield from redis.ping()


@pytest.mark.run_loop
def test_select(redis):
    assert redis.db == 0

    resp = yield from redis.select(1)
    assert resp is True
    assert redis.db == 1
    assert redis.connection.db == 1


@pytest.mark.run_loop
def test_encoding(create_redis, loop, server):
    redis = yield from create_redis(
        server.tcp_address,
        db=1, encoding='utf-8',
        loop=loop)
    assert redis.encoding == 'utf-8'
